# Errors.


class FionaValueError(ValueError):
    """Fiona-specific value errors"""


class DriverError(FionaValueError):
    """Encapsulates unsupported driver and driver mode errors."""


class SchemaError(FionaValueError):
    """When a schema mapping has no properties or no geometry."""


class CRSError(FionaValueError):
    """When a crs mapping has neither init or proj items."""


class DataIOError(IOError):
    """IO errors involving driver registration or availability."""


class DriverIOError(IOError):
    """A format specific driver error."""


class FieldNameEncodeError(UnicodeEncodeError):
    """Failure to encode a field name."""


class UnsupportedGeometryTypeError(KeyError):
    """When a OGR geometry type isn't supported by Fiona."""



# Test of opening and closing and opening

import logging
import os.path
import shutil
import subprocess
import sys
import tempfile
import unittest

import fiona

logging.basicConfig(stream=sys.stderr, level=logging.INFO)
log = logging.getLogger('fiona.tests')

class RevolvingDoorTest(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
    
    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_write_revolving_door(self):

        with fiona.open('tests/data/coutwildrnp.shp') as src:
            meta = src.meta
            features = list(src)

        shpname = os.path.join(self.tempdir, 'foo.shp')
        
        with fiona.open(shpname, 'w', **meta) as dst:
            dst.writerecords(features)

        with fiona.open(shpname) as src:
            pass



import json
import sys
import unittest

from click.testing import CliRunner

from fiona.fio import cat

from .fixtures import feature_seq
from .fixtures import feature_seq_pp_rs


WILDSHP = 'tests/data/coutwildrnp.shp'

FIXME_WINDOWS = sys.platform.startswith('win')

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_one():
    runner = CliRunner()
    result = runner.invoke(cat.cat, [WILDSHP])
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 67

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_two():
    runner = CliRunner()
    result = runner.invoke(cat.cat, [WILDSHP, WILDSHP])
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 134

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_bbox_no():
    runner = CliRunner()
    result = runner.invoke(
        cat.cat,
        [WILDSHP, '--bbox', '0,10,80,20'],
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output == ""

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_bbox_yes():
    runner = CliRunner()
    result = runner.invoke(
        cat.cat,
        [WILDSHP, '--bbox', '-109,37,-107,39'],
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 19

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_bbox_json_yes():
    runner = CliRunner()
    result = runner.invoke(
        cat.cat,
        [WILDSHP, '--bbox', '[-109,37,-107,39]'],
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 19


def test_multi_layer():
    layerdef = "1:coutwildrnp,1:coutwildrnp"
    runner = CliRunner()
    result = runner.invoke(
        cat.cat, ['--layer', layerdef, 'tests/data/'])
    assert result.output.count('"Feature"') == 134


def test_multi_layer_fail():
    runner = CliRunner()
    result = runner.invoke(cat.cat, ['--layer',
                           '200000:coutlildrnp',
                           'tests/data'])
    assert result.exit_code != 0



from __future__ import division
import json

from click.testing import CliRunner

from fiona.fio.calc import calc
from .fixtures import feature_seq


def test_fail():
    runner = CliRunner()
    result = runner.invoke(calc,
                           ["TEST", "f.properties.test > 5"],
                           '{"type": "no_properties"}')
    assert result.exit_code == 1


def _load(output):
    features = []
    for x in output.splitlines():
        try:
            features.append(json.loads(x))
        except:
            pass  # nosetests puts some debugging garbage to stdout
    return features


def test_calc_seq():
    runner = CliRunner()

    result = runner.invoke(calc,
                           ["TEST", "f.properties.AREA / f.properties.PERIMETER"],
                           feature_seq)
    assert result.exit_code == 0

    feats = _load(result.output)
    assert len(feats) == 2
    for feat in feats:
        assert feat['properties']['TEST'] == \
            feat['properties']['AREA'] / feat['properties']['PERIMETER']


def test_bool_seq():
    runner = CliRunner()

    result = runner.invoke(calc,
                           ["TEST", "f.properties.AREA > 0.015"],
                           feature_seq)
    assert result.exit_code == 0
    feats = _load(result.output)
    assert len(feats) == 2
    assert feats[0]['properties']['TEST'] == True
    assert feats[1]['properties']['TEST'] == False


def test_existing_property():
    runner = CliRunner()

    result = runner.invoke(calc,
                           ["AREA", "f.properties.AREA * 2"],
                           feature_seq)
    assert result.exit_code == 1

    result = runner.invoke(calc,
                           ["--overwrite", "AREA", "f.properties.AREA * 2"],
                           feature_seq)
    assert result.exit_code == 0
    feats = _load(result.output)
    assert len(feats) == 2
    for feat in feats:
        assert 'AREA' in feat['properties']



# Testing BytesCollection
import sys
import unittest

import pytest
import six

import fiona

FIXME_WINDOWS = sys.platform.startswith('win')


class ReadingTest(unittest.TestCase):

    def setUp(self):
        with open('tests/data/coutwildrnp.json') as src:
            bytesbuf = src.read().encode('utf-8')
        self.c = fiona.BytesCollection(bytesbuf)

    def tearDown(self):
        self.c.close()

    @unittest.skipIf(six.PY2, 'string are bytes in Python 2')
    def test_construct_with_str(self):
        with open('tests/data/coutwildrnp.json') as src:
            strbuf = src.read()
        self.assertRaises(ValueError, fiona.BytesCollection, strbuf)

    @unittest.skipIf(FIXME_WINDOWS,
                     reason="FIXME on Windows. Please look into why this test is not working.")
    def test_open_repr(self):
        # I'm skipping checking the name of the virtual file as it produced by uuid.
        self.assertTrue(repr(self.c).startswith("<open BytesCollection '/vsimem/"))

    @unittest.skipIf(FIXME_WINDOWS,
                     reason="FIXME on Windows. Please look into why this test is not working.")
    def test_closed_repr(self):
        # I'm skipping checking the name of the virtual file as it produced by uuid.
        self.c.close()
        self.assertTrue(repr(self.c).startswith("<closed BytesCollection '/vsimem/"))

    def test_path(self):
        self.assertEqual(self.c.path, self.c.virtual_file)

    def test_closed_virtual_file(self):
        self.c.close()
        self.assertTrue(self.c.virtual_file is None)

    def test_closed_buf(self):
        self.c.close()
        self.assertTrue(self.c.bytesbuf is None)

    def test_name(self):
        self.assertTrue(len(self.c.name) > 0)

    def test_mode(self):
        self.assertEqual(self.c.mode, 'r')

    @unittest.skipIf(FIXME_WINDOWS,
                     reason="FIXME on Windows. Please look into why this test is not working.")
    def test_collection(self):
        self.assertEqual(self.c.encoding, 'utf-8')

    def test_iter(self):
        self.assertTrue(iter(self.c))

    def test_closed_no_iter(self):
        self.c.close()
        self.assertRaises(ValueError, iter, self.c)

    def test_len(self):
        self.assertEqual(len(self.c), 67)

    def test_closed_len(self):
        # Len is lazy, it's never computed in this case. TODO?
        self.c.close()
        self.assertEqual(len(self.c), 0)

    def test_len_closed_len(self):
        # Lazy len is computed in this case and sticks.
        len(self.c)
        self.c.close()
        self.assertEqual(len(self.c), 67)

    def test_driver(self):
        self.assertEqual(self.c.driver, "GeoJSON")

    def test_closed_driver(self):
        self.c.close()
        self.assertEqual(self.c.driver, None)

    def test_driver_closed_driver(self):
        self.c.driver
        self.c.close()
        self.assertEqual(self.c.driver, "GeoJSON")

    def test_schema(self):
        s = self.c.schema['properties']
        self.assertEqual(s['PERIMETER'], "float")
        self.assertEqual(s['NAME'], "str")
        self.assertEqual(s['URL'], "str")
        self.assertEqual(s['STATE_FIPS'], "str")
        self.assertEqual(s['WILDRNP020'], "int")

    def test_closed_schema(self):
        # Schema is lazy too, never computed in this case. TODO?
        self.c.close()
        self.assertEqual(self.c.schema, None)

    def test_schema_closed_schema(self):
        self.c.schema
        self.c.close()
        self.assertEqual(
            sorted(self.c.schema.keys()),
            ['geometry', 'properties'])

    def test_crs(self):
        crs = self.c.crs
        self.assertEqual(crs['init'], 'epsg:4326')

    def test_crs_wkt(self):
        crs = self.c.crs_wkt
        self.assertTrue(crs.startswith('GEOGCS["WGS 84"'))

    def test_closed_crs(self):
        # Crs is lazy too, never computed in this case. TODO?
        self.c.close()
        self.assertEqual(self.c.crs, None)

    def test_crs_closed_crs(self):
        self.c.crs
        self.c.close()
        self.assertEqual(
            sorted(self.c.crs.keys()),
            ['init'])

    def test_meta(self):
        self.assertEqual(
            sorted(self.c.meta.keys()),
            ['crs', 'crs_wkt', 'driver', 'schema'])

    def test_bounds(self):
        self.assertAlmostEqual(self.c.bounds[0], -113.564247, 6)
        self.assertAlmostEqual(self.c.bounds[1], 37.068981, 6)
        self.assertAlmostEqual(self.c.bounds[2], -104.970871, 6)
        self.assertAlmostEqual(self.c.bounds[3], 41.996277, 6)

    def test_iter_one(self):
        itr = iter(self.c)
        f = next(itr)
        self.assertEqual(f['id'], "0")
        self.assertEqual(f['properties']['STATE'], 'UT')

    def test_iter_list(self):
        f = list(self.c)[0]
        self.assertEqual(f['id'], "0")
        self.assertEqual(f['properties']['STATE'], 'UT')

    def test_re_iter_list(self):
        f = list(self.c)[0]  # Run through iterator
        f = list(self.c)[0]  # Run through a new, reset iterator
        self.assertEqual(f['id'], "0")
        self.assertEqual(f['properties']['STATE'], 'UT')

    def test_getitem_one(self):
        f = self.c[0]
        self.assertEqual(f['id'], "0")
        self.assertEqual(f['properties']['STATE'], 'UT')

    def test_no_write(self):
        self.assertRaises(IOError, self.c.write, {})

    def test_iter_items_list(self):
        i, f = list(self.c.items())[0]
        self.assertEqual(i, 0)
        self.assertEqual(f['id'], "0")
        self.assertEqual(f['properties']['STATE'], 'UT')

    def test_iter_keys_list(self):
        i = list(self.c.keys())[0]
        self.assertEqual(i, 0)

    def test_in_keys(self):
        self.assertTrue(0 in self.c.keys())
        self.assertTrue(0 in self.c)

class FilterReadingTest(unittest.TestCase):

    def setUp(self):
        with open('tests/data/coutwildrnp.json') as src:
            bytesbuf = src.read().encode('utf-8')
        self.c = fiona.BytesCollection(bytesbuf)

    def tearDown(self):
        self.c.close()

    def test_filter_1(self):
        results = list(self.c.filter(bbox=(-120.0, 30.0, -100.0, 50.0)))
        self.assertEqual(len(results), 67)
        f = results[0]
        self.assertEqual(f['id'], "0")
        self.assertEqual(f['properties']['STATE'], 'UT')

    def test_filter_reset(self):
        results = list(self.c.filter(bbox=(-112.0, 38.0, -106.0, 40.0)))
        self.assertEqual(len(results), 26)
        results = list(self.c.filter())
        self.assertEqual(len(results), 67)

    def test_filter_mask(self):
        mask = {
            'type': 'Polygon',
            'coordinates': (
                ((-112, 38), (-112, 40), (-106, 40), (-106, 38), (-112, 38)),)}
        results = list(self.c.filter(mask=mask))
        self.assertEqual(len(results), 26)


def test_zipped_bytes_collection():
    with open('tests/data/coutwildrnp.zip', 'rb') as src:
        zip_file_bytes = src.read()

    with fiona.BytesCollection(zip_file_bytes) as col:
        assert col.name == 'coutwildrnp'


def test_grenada_bytes_geojson():
    """Read grenada.geojson as BytesCollection.

    grenada.geojson is an example of geojson that GDAL's GeoJSON
    driver will fail to read successfully unless the file's extension
    reflects its json'ness.
    """
    with open('tests/data/grenada.geojson', 'rb') as src:
        bytes_grenada_geojson = src.read()
    
    # We expect an exception if the GeoJSON driver isn't specified.
    with pytest.raises(fiona.errors.FionaValueError):
        with fiona.BytesCollection(bytes_grenada_geojson) as col:
            pass

    # If told what driver to use, we should be good.
    with fiona.BytesCollection(bytes_grenada_geojson, driver='GeoJSON') as col:        
        assert len(col) == 1
        



import json
from pkg_resources import iter_entry_points
import re
import sys
import unittest

from click.testing import CliRunner

from fiona.fio.main import main_group


WILDSHP = 'tests/data/coutwildrnp.shp'

FIXME_WINDOWS = sys.platform.startswith('win')

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_info_json():
    runner = CliRunner()
    result = runner.invoke(main_group, ['info', WILDSHP])
    assert result.exit_code == 0
    assert '"count": 67' in result.output
    assert '"crs": "EPSG:4326"' in result.output
    assert '"driver": "ESRI Shapefile"' in result.output
    assert '"name": "coutwildrnp"' in result.output

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_info_count():
    runner = CliRunner()
    result = runner.invoke(main_group, ['info', '--count', WILDSHP])
    assert result.exit_code == 0
    assert result.output == "67\n"

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_info_bounds():
    runner = CliRunner()
    result = runner.invoke(main_group, ['info', '--bounds', WILDSHP])
    assert result.exit_code == 0
    assert len(re.findall(r'\d*\.\d*', result.output)) == 4


def test_all_registered():
    # Make sure all the subcommands are actually registered to the main CLI group
    for ep in iter_entry_points('fiona.fio_commands'):
        assert ep.name in main_group.commands


def _filter_info_warning(lines):
    """$ fio info can issue a RuntimeWarning, but click adds stderr to stdout
    so we have to filter it out before decoding JSON lines."""
    lines = list(filter(lambda x: 'RuntimeWarning' not in x, lines))
    return lines


@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_info_no_count():
    """Make sure we can still get a `$ fio info` report on datasources that do
    not support feature counting, AKA `len(collection)`.
    """
    runner = CliRunner()
    result = runner.invoke(main_group, ['info', 'tests/data/test_gpx.gpx'])
    assert result.exit_code == 0
    lines = _filter_info_warning(result.output.splitlines())
    assert len(lines) == 1, "First line is warning & second is JSON.  No more."
    assert json.loads(lines[0])['count'] is None


@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_info_layer():
    for layer in ('routes', '1'):
        runner = CliRunner()
        result = runner.invoke(main_group, [
            'info',
            'tests/data/test_gpx.gpx',
            '--layer', layer])
        print(result.output)
        assert result.exit_code == 0
        lines = _filter_info_warning(result.output.splitlines())
        assert len(lines) == 1, "1st line is warning & 2nd is JSON - no more."
        assert json.loads(lines[0])['name'] == 'routes'



"""$ fio dump"""


from functools import partial
import json
import logging

import click
import cligj

import fiona
from fiona.fio import helpers
from fiona.fio import options
from fiona.transform import transform_geom


@click.command(short_help="Dump a dataset to GeoJSON.")
@click.argument('input', type=click.Path(), required=True)
@click.option('--layer', metavar="INDEX|NAME", callback=options.cb_layer,
              help="Print information about a specific layer.  The first "
                   "layer is used by default.  Layers use zero-based "
                   "numbering when accessed by index.")
@click.option('--encoding', help="Specify encoding of the input file.")
@cligj.precision_opt
@cligj.indent_opt
@cligj.compact_opt
@click.option('--record-buffered/--no-record-buffered', default=False,
              help="Economical buffering of writes at record, not collection "
                   "(default), level.")
@click.option('--ignore-errors/--no-ignore-errors', default=False,
              help="log errors but do not stop serialization.")
@click.option('--with-ld-context/--without-ld-context', default=False,
              help="add a JSON-LD context to JSON output.")
@click.option('--add-ld-context-item', multiple=True,
              help="map a term to a URI and add it to the output's JSON LD "
                   "context.")
@click.pass_context
def dump(ctx, input, encoding, precision, indent, compact, record_buffered,
         ignore_errors, with_ld_context, add_ld_context_item, layer):

    """Dump a dataset either as a GeoJSON feature collection (the default)
    or a sequence of GeoJSON features."""

    verbosity = (ctx.obj and ctx.obj['verbosity']) or 2
    logger = logging.getLogger('fio')
    sink = click.get_text_stream('stdout')

    dump_kwds = {'sort_keys': True}
    if indent:
        dump_kwds['indent'] = indent
    if compact:
        dump_kwds['separators'] = (',', ':')
    item_sep = compact and ',' or ', '

    open_kwds = {}
    if encoding:
        open_kwds['encoding'] = encoding
    if layer:
        open_kwds['layer'] = layer

    def transformer(crs, feat):
        tg = partial(transform_geom, crs, 'EPSG:4326',
                     antimeridian_cutting=True, precision=precision)
        feat['geometry'] = tg(feat['geometry'])
        return feat

    try:
        with fiona.drivers(CPL_DEBUG=verbosity > 2):
            with fiona.open(input, **open_kwds) as source:
                meta = source.meta
                meta['fields'] = dict(source.schema['properties'].items())

                if record_buffered:
                    # Buffer GeoJSON data at the feature level for smaller
                    # memory footprint.
                    indented = bool(indent)
                    rec_indent = "\n" + " " * (2 * (indent or 0))

                    collection = {
                        'type': 'FeatureCollection',
                        'fiona:schema': meta['schema'],
                        'fiona:crs': meta['crs'],
                        'features': []}
                    if with_ld_context:
                        collection['@context'] = helpers.make_ld_context(
                            add_ld_context_item)

                    head, tail = json.dumps(
                        collection, **dump_kwds).split('[]')

                    sink.write(head)
                    sink.write("[")

                    itr = iter(source)

                    # Try the first record.
                    try:
                        i, first = 0, next(itr)
                        first = transformer(first)
                        if with_ld_context:
                            first = helpers.id_record(first)
                        if indented:
                            sink.write(rec_indent)
                        sink.write(json.dumps(
                            first, **dump_kwds).replace("\n", rec_indent))
                    except StopIteration:
                        pass
                    except Exception as exc:
                        # Ignoring errors is *not* the default.
                        if ignore_errors:
                            logger.error(
                                "failed to serialize file record %d (%s), "
                                "continuing",
                                i, exc)
                        else:
                            # Log error and close up the GeoJSON, leaving it
                            # more or less valid no matter what happens above.
                            logger.critical(
                                "failed to serialize file record %d (%s), "
                                "quiting",
                                i, exc)
                            sink.write("]")
                            sink.write(tail)
                            if indented:
                                sink.write("\n")
                            raise

                    # Because trailing commas aren't valid in JSON arrays
                    # we'll write the item separator before each of the
                    # remaining features.
                    for i, rec in enumerate(itr, 1):
                        rec = transformer(rec)
                        try:
                            if with_ld_context:
                                rec = helpers.id_record(rec)
                            if indented:
                                sink.write(rec_indent)
                            sink.write(item_sep)
                            sink.write(json.dumps(
                                rec, **dump_kwds).replace("\n", rec_indent))
                        except Exception as exc:
                            if ignore_errors:
                                logger.error(
                                    "failed to serialize file record %d (%s), "
                                    "continuing",
                                    i, exc)
                            else:
                                logger.critical(
                                    "failed to serialize file record %d (%s), "
                                    "quiting",
                                    i, exc)
                                sink.write("]")
                                sink.write(tail)
                                if indented:
                                    sink.write("\n")
                                raise

                    # Close up the GeoJSON after writing all features.
                    sink.write("]")
                    sink.write(tail)
                    if indented:
                        sink.write("\n")

                else:
                    # Buffer GeoJSON data at the collection level. The default.
                    collection = {
                        'type': 'FeatureCollection',
                        'fiona:schema': meta['schema'],
                        'fiona:crs': meta['crs']}
                    if with_ld_context:
                        collection['@context'] = helpers.make_ld_context(
                            add_ld_context_item)
                        collection['features'] = [
                            helpers.id_record(transformer(rec))
                            for rec in source]
                    else:
                        collection['features'] = [
                            transformer(source.crs, rec) for rec in source]
                    json.dump(collection, sink, **dump_kwds)

    except Exception:
        logger.exception("Exception caught during processing")
        raise click.Abort()



import logging
import os
import shutil
import sys
import tempfile
import unittest

import fiona
from fiona.compat import OrderedDict

logging.basicConfig(stream=sys.stderr, level=logging.INFO)

FIXME_WINDOWS = sys.platform.startswith("win")

class ReadAccess(unittest.TestCase):
    # To check that we'll be able to get multiple 'r' connections to layers
    # in a single file.
    
    def setUp(self):
        self.c = fiona.open("tests/data/coutwildrnp.shp", "r", layer="coutwildrnp")
    
    def tearDown(self):
        self.c.close()

    def test_meta(self):
        with fiona.open("tests/data/coutwildrnp.shp", "r", layer="coutwildrnp") as c2:
            self.assertEqual(len(self.c), len(c2))
            self.assertEqual(sorted(self.c.schema.items()), sorted(c2.schema.items()))

    def test_meta(self):
        f1 = next(iter(self.c))
        with fiona.open("tests/data/coutwildrnp.shp", "r", layer="coutwildrnp") as c2:
            f2 = next(iter(c2))
            self.assertEqual(f1, f2)

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. These tests raise PermissionErrors on Windows in Python 3.x (which doesn't exist in Python 2.7).  Please look into why this test isn't working.")
class ReadWriteAccess(unittest.TestCase):
    # To check that we'll be able to read from a file that we're
    # writing to.
    
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.c = fiona.open(
            os.path.join(self.tempdir, "multi_write_test.shp"),
            "w",
            driver="ESRI Shapefile",
            schema={
                'geometry': 'Point', 
                'properties': [('title', 'str:80'), ('date', 'date')]},
            crs={'init': "epsg:4326", 'no_defs': True},
            encoding='utf-8')
        self.f = {
            'type': 'Feature',
            'geometry': {'type': 'Point', 'coordinates': (0.0, 0.1)},
            'properties': OrderedDict([('title', 'point one'), ('date', '2012-01-29')])}
        self.c.writerecords([self.f])
        self.c.flush()

    def tearDown(self):
        self.c.close()
        shutil.rmtree(self.tempdir)

    def test_meta(self):
        c2 = fiona.open(os.path.join(self.tempdir, "multi_write_test.shp"), "r")
        self.assertEqual(len(self.c), len(c2))
        self.assertEqual(sorted(self.c.schema.items()), sorted(c2.schema.items()))

    def test_read(self):
        c2 = fiona.open(os.path.join(self.tempdir, "multi_write_test.shp"), "r")
        f2 = next(iter(c2))
        del f2['id']
        self.assertEqual(self.f, f2)

    def test_read_after_close(self):
        c2 = fiona.open(os.path.join(self.tempdir, "multi_write_test.shp"), "r")
        self.c.close()
        f2 = next(iter(c2))
        del f2['id']
        self.assertEqual(self.f, f2)

@unittest.skipIf(FIXME_WINDOWS, 
                    reason="FIXME on Windows. These tests raise PermissionErrors on Windows in Python 3.x (which doesn't exist in Python 2.7).  Please look into why this test isn't working.")
class LayerCreation(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.dir = os.path.join(self.tempdir, 'layer_creation')
        if os.path.exists(self.dir):
            shutil.rmtree(self.dir)
        os.mkdir(self.dir)
        self.c = fiona.open(
            self.dir,
            'w',
            layer='write_test',
            driver='ESRI Shapefile',
            schema={
                'geometry': 'Point', 
                'properties': [('title', 'str:80'), ('date', 'date')]},
            crs={'init': "epsg:4326", 'no_defs': True},
            encoding='utf-8')
        self.f = {
            'type': 'Feature',
            'geometry': {'type': 'Point', 'coordinates': (0.0, 0.1)},
            'properties': OrderedDict([('title', 'point one'), ('date', '2012-01-29')])}
        self.c.writerecords([self.f])
        self.c.flush()

    def tearDown(self):
        self.c.close()
        shutil.rmtree(self.tempdir)

    def test_meta(self):
        c2 = fiona.open(os.path.join(self.dir, "write_test.shp"), "r")
        self.assertEqual(len(self.c), len(c2))
        self.assertEqual(sorted(self.c.schema.items()), sorted(c2.schema.items()))

    def test_read(self):
        c2 = fiona.open(os.path.join(self.dir, "write_test.shp"), "r")
        f2 = next(iter(c2))
        del f2['id']
        self.assertEqual(self.f, f2)

    def test_read_after_close(self):
        c2 = fiona.open(os.path.join(self.dir, "write_test.shp"), "r")
        self.c.close()
        f2 = next(iter(c2))
        del f2['id']
        self.assertEqual(self.f, f2)



"""$ fio distrib"""


import json
import logging

import click
import cligj

from fiona.fio import helpers


@click.command()
@cligj.use_rs_opt
@click.pass_context
def distrib(ctx, use_rs):

    """Distribute features from a collection.

    Print the features of GeoJSON objects read from stdin.
    """

    verbosity = (ctx.obj and ctx.obj['verbosity']) or 2
    logger = logging.getLogger('fio')
    stdin = click.get_text_stream('stdin')
    try:
        source = helpers.obj_gen(stdin)
        for i, obj in enumerate(source):
            obj_id = obj.get('id', 'collection:' + str(i))
            features = obj.get('features') or [obj]
            for j, feat in enumerate(features):
                if obj.get('type') == 'FeatureCollection':
                    feat['parent'] = obj_id
                feat_id = feat.get('id', 'feature:' + str(i))
                feat['id'] = feat_id
                if use_rs:
                    click.echo(u'\u001e', nl=False)
                click.echo(json.dumps(feat))
    except Exception:
        logger.exception("Exception caught during processing")
        raise click.Abort()



"""Unittests for $ fio dump"""

import sys
import unittest

from click.testing import CliRunner

from fiona.fio import dump


WILDSHP = 'tests/data/coutwildrnp.shp'
TESTGPX = 'tests/data/test_gpx.gpx'

FIXME_WINDOWS = sys.platform.startswith('win')

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_dump():
    runner = CliRunner()
    result = runner.invoke(dump.dump, [WILDSHP])
    assert result.exit_code == 0
    assert '"FeatureCollection"' in result.output


def test_dump_layer():
    for layer in ('routes', '1'):
        runner = CliRunner()
        result = runner.invoke(dump.dump, [TESTGPX, '--layer', layer])
        assert result.exit_code == 0
        assert '"FeatureCollection"' in result.output



import json
import os.path
from six import text_type
import tempfile

import fiona
from fiona import prop_type, prop_width
from fiona.rfc3339 import FionaDateType


def test_width_str():
    assert prop_width('str:254') == 254
    assert prop_width('str') == 80


def test_width_other():
    assert prop_width('int') == None
    assert prop_width('float') == None
    assert prop_width('date') == None


def test_types():
    assert prop_type('str:254') == text_type
    assert prop_type('str') == text_type
    assert isinstance(0, prop_type('int'))
    assert isinstance(0.0, prop_type('float'))
    assert prop_type('date') == FionaDateType


def test_read_json_object_properties():
    """JSON object properties are properly serialized"""
    data = """
{
  "type": "FeatureCollection",
  "features": [
    {
      "geometry": {
        "type": "Polygon",
        "coordinates": [
          [
            [
              87.33588,
              43.53139
            ],
            [
              87.33588,
              45.66894
            ],
            [
              90.27542,
              45.66894
            ],
            [
              90.27542,
              43.53139
            ],
            [
              87.33588,
              43.53139
            ]
          ]
        ]
      },
      "type": "Feature",
      "properties": {
        "upperLeftCoordinate": {
          "latitude": 45.66894,
          "longitude": 87.91166
        },
        "tricky": "{gotcha"
      }
    }
  ]
}
"""
    tmpdir = tempfile.mkdtemp()
    filename = os.path.join(tmpdir, 'test.json')

    with open(filename, 'w') as f:
        f.write(data)

    with fiona.open(filename) as src:
        ftr = next(iter(src))
        props = ftr['properties']
        assert props['upperLeftCoordinate']['latitude'] == 45.66894
        assert props['upperLeftCoordinate']['longitude'] == 87.91166
        assert props['tricky'] == "{gotcha"


def test_write_json_object_properties():
    """Python object properties are properly serialized"""
    data = """
{
  "type": "FeatureCollection",
  "features": [
    {
      "geometry": {
        "type": "Polygon",
        "coordinates": [
          [
            [
              87.33588,
              43.53139
            ],
            [
              87.33588,
              45.66894
            ],
            [
              90.27542,
              45.66894
            ],
            [
              90.27542,
              43.53139
            ],
            [
              87.33588,
              43.53139
            ]
          ]
        ]
      },
      "type": "Feature",
      "properties": {
        "upperLeftCoordinate": {
          "latitude": 45.66894,
          "longitude": 87.91166
        },
        "tricky": "{gotcha"
      }
    }
  ]
}
"""
    data = json.loads(data)['features'][0]
    tmpdir = tempfile.mkdtemp()
    filename = os.path.join(tmpdir, 'test.json')
    with fiona.open(
            filename, 'w',
            driver='GeoJSON',
            schema={
                'geometry': 'Polygon',
                'properties': {'upperLeftCoordinate': 'str', 'tricky': 'str'}}
            ) as dst:
        dst.write(data)

    with fiona.open(filename) as src:
        ftr = next(iter(src))
        props = ftr['properties']
        assert props['upperLeftCoordinate']['latitude'] == 45.66894
        assert props['upperLeftCoordinate']['longitude'] == 87.91166
        assert props['tricky'] == "{gotcha"


def test_json_prop_decode_non_geojson_driver():
    feature = {
        "type": "Feature",
        "properties": {
            "ulc": {
                "latitude": 45.66894,
                "longitude": 87.91166
            },
            "tricky": "{gotcha"
        },
        "geometry": {
            "type": "Point",
            "coordinates": [10, 15]
        }
    }

    meta = {
        'crs': 'EPSG:4326',
        'driver': 'ESRI Shapefile',
        'schema': {
            'geometry': 'Point',
            'properties': {
                'ulc': 'str:255',
                'tricky': 'str:255'
            }
        }
    }

    tmpdir = tempfile.mkdtemp()
    filename = os.path.join(tmpdir, 'test.json')
    with fiona.open(filename, 'w', **meta) as dst:
        dst.write(feature)

    with fiona.open(filename) as src:
        actual = next(iter(src))

    assert isinstance(actual['properties']['ulc'], text_type)
    a = json.loads(actual['properties']['ulc'])
    e = json.loads(actual['properties']['ulc'])
    assert e == a
    assert actual['properties']['tricky'].startswith('{')



# testing Fiona's RFC 3339 support, to be called by nosetests

import logging
import re
import sys
import unittest

from fiona.rfc3339 import parse_date, parse_datetime, parse_time
from fiona.rfc3339 import group_accessor, pattern_date

logging.basicConfig(stream=sys.stderr, level=logging.INFO)

class DateParseTest(unittest.TestCase):

    def test_yyyymmdd(self):
        self.assertEqual(
            parse_date("2012-01-29"), (2012, 1, 29, 0, 0, 0, 0.0))

    def test_error(self):
        self.assertRaises(ValueError, parse_date, ("xxx"))

class TimeParseTest(unittest.TestCase):
    
    def test_hhmmss(self):
        self.assertEqual(
            parse_time("10:11:12"), (0, 0, 0, 10, 11, 12, 0.0))

    def test_hhmm(self):
        self.assertEqual(
            parse_time("10:11"), (0, 0, 0, 10, 11, 0, 0.0))

    def test_hhmmssff(self):
        self.assertEqual(
            parse_time("10:11:12.42"), 
            (0, 0, 0, 10, 11, 12, 0.42*1000000.0))

    def test_hhmmssz(self):
        self.assertEqual(
            parse_time("10:11:12Z"), (0, 0, 0, 10, 11, 12, 0.0))

    def test_hhmmssoff(self):
        self.assertEqual(
            parse_time("10:11:12-01:00"), (0, 0, 0, 10, 11, 12, 0.0))

    def test_error(self):
        self.assertRaises(ValueError, parse_time, ("xxx"))

class DatetimeParseTest(unittest.TestCase):
    
    def test_yyyymmdd(self):
        self.assertEqual(
            parse_datetime("2012-01-29T10:11:12"), 
            (2012, 1, 29, 10, 11, 12, 0.0))

    def test_error(self):
        self.assertRaises(ValueError, parse_datetime, ("xxx"))

def test_group_accessor_indexerror():
    match = re.search(pattern_date, '2012-01-29')
    g = group_accessor(match)
    assert g.group(-1) == 0
    assert g.group(6) == 0




# testing features, to be called by nosetests

import logging
import os
import shutil
import sys
import tempfile
import unittest

from fiona import collection
from fiona.collection import Collection
from fiona.ogrext import featureRT

#logging.basicConfig(stream=sys.stderr, level=logging.INFO)

class PointRoundTripTest(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        schema = {'geometry': 'Point', 'properties': {'title': 'str'}}
        self.c = Collection(os.path.join(self.tempdir, "foo.shp"),
                            "w", driver="ESRI Shapefile", schema=schema)
    def tearDown(self):
        self.c.close()
        shutil.rmtree(self.tempdir)
    def test_geometry(self):
        f = { 'id': '1', 
              'geometry': {'type': 'Point', 'coordinates': (0.0, 0.0)},
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        self.assertEqual(
            sorted(g['geometry'].items()),
            [('coordinates', (0.0, 0.0)), ('type', 'Point')])
    def test_properties(self):
        f = { 'id': '1', 
              'geometry': {'type': 'Point', 'coordinates': (0.0, 0.0)},
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        self.assertEqual(g['properties']['title'], 'foo')
    def test_none_property(self):
        f = { 'id': '1',
              'geometry': {'type': 'Point', 'coordinates': (0.0, 0.0)},
              'properties': {'title': None} }
        g = featureRT(f, self.c)
        self.assertEqual(g['properties']['title'], None)

class LineStringRoundTripTest(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        schema = {'geometry': 'LineString', 'properties': {'title': 'str'}}
        self.c = Collection(os.path.join(self.tempdir, "foo.shp"),
                            "w", "ESRI Shapefile", schema=schema)
    def tearDown(self):
        self.c.close()
        shutil.rmtree(self.tempdir)
    def test_geometry(self):
        f = { 'id': '1', 
              'geometry': { 'type': 'LineString', 
                            'coordinates': [(0.0, 0.0), (1.0, 1.0)] },
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        self.assertEqual(
            sorted(g['geometry'].items()),
            [('coordinates', [(0.0, 0.0), (1.0, 1.0)]), 
             ('type', 'LineString')])
    def test_properties(self):
        f = { 'id': '1',
              'geometry': {'type': 'Point', 'coordinates': (0.0, 0.0)},
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        self.assertEqual(g['properties']['title'], 'foo')

class PolygonRoundTripTest(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        schema = {'geometry': 'Polygon', 'properties': {'title': 'str'}}
        self.c = Collection(os.path.join(self.tempdir, "foo.shp"),
                            "w", "ESRI Shapefile", schema=schema)
    def tearDown(self):
        self.c.close()
        shutil.rmtree(self.tempdir)
    def test_geometry(self):
        f = { 'id': '1', 
              'geometry': { 'type': 'Polygon', 
                            'coordinates': 
                                [[(0.0, 0.0), 
                                  (0.0, 1.0), 
                                  (1.0, 1.0), 
                                  (1.0, 0.0), 
                                  (0.0, 0.0)]] },
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        self.assertEqual(
            sorted(g['geometry'].items()),
            [('coordinates', [[(0.0, 0.0), 
                                  (0.0, 1.0), 
                                  (1.0, 1.0), 
                                  (1.0, 0.0), 
                                  (0.0, 0.0)]]), 
             ('type', 'Polygon')])
    def test_properties(self):
        f = { 'id': '1', 
              'geometry': { 'type': 'Polygon', 
                            'coordinates': 
                                [[(0.0, 0.0), 
                                  (0.0, 1.0), 
                                  (1.0, 1.0), 
                                  (1.0, 0.0), 
                                  (0.0, 0.0)]] },
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        self.assertEqual(g['properties']['title'], 'foo')

